@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2010 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -150,6 +150,9 @@ const TCHAR *wrapperGetJState(int jState) {
     case WRAPPER_JSTATE_DOWN_CHECK:
         name = TEXT("DOWN_CHECK");
         break;
+    case WRAPPER_JSTATE_DOWN_FLUSH:
+        name = TEXT("DOWN_FLUSH");
+        break;
     default:
         name = TEXT("UNKNOWN");
         break;
@@ -401,12 +404,13 @@ void anchorPoll(TICKS nowTicks) {
                     (wrapperData->jState == WRAPPER_JSTATE_KILLING) ||
                     (wrapperData->jState == WRAPPER_JSTATE_KILL) ||
                     (wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK) ||
+                    (wrapperData->jState == WRAPPER_JSTATE_DOWN_FLUSH) ||
                     (wrapperData->jState == WRAPPER_JSTATE_DOWN_CLEAN)) {
                     /* Already shutting down, so nothing more to do. */
                 } else {
-                    wrapperStopProcess(0);
+                    /* Always force the shutdown as this was an external event. */
+                    wrapperStopProcess(0, TRUE);
                 }
-                
 
                 /* To make sure that the JVM will not be restarted for any reason,
                  *  start the Wrapper shutdown process as well. */
@@ -447,8 +451,10 @@ void commandPoll(TICKS nowTicks) {
     TCHAR *c;
     TCHAR *d;
     TCHAR *command;
-    TCHAR *params;
+    TCHAR *param1;
+    TCHAR *param2;
     int exitCode;
+    int pauseTime;
     int logLevel;
     int oldLowLogLevel;
     int newLowLogLevel;
@@ -507,14 +513,34 @@ void commandPoll(TICKS nowTicks) {
 
                             command = buffer;
 
-                            /** Look for the first space, everything after it will be the parameter. */
-                            if ((params = _tcschr(buffer, ' ')) != NULL ) {
-                                params[0] = TEXT('\0');
+                            /** Look for the first space, everything after it will be the parameter(s). */
+                            /* Look for parameter 1. */
+                            if ((param1 = _tcschr(buffer, ' ')) != NULL ) {
+                                param1[0] = TEXT('\0'); /* Terminate the command. */
 
                                 /* Find the first non-space character. */
                                 do {
-                                    params++;
-                                } while (params[0] == TEXT(' '));
+                                    param1++;
+                                } while (param1[0] == TEXT(' '));
+                            }
+                            if (param1 != NULL) {
+                                /* Look for parameter 2. */
+                                if ((param2 = _tcschr(param1, ' ')) != NULL ) {
+                                    param2[0] = TEXT('\0'); /* Terminate param1. */
+    
+                                    /* Find the first non-space character. */
+                                    do {
+                                        param2++;
+                                    } while (param2[0] == TEXT(' '));
+                                }
+                                if (param2 != NULL) {
+                                    /* Make sure parameter 2 is terminated. */
+                                    if ((d = _tcschr(param2, ' ')) != NULL ) {
+                                        d[0] = TEXT('\0'); /* Terminate param2. */
+                                    }
+                                }
+                            } else {
+                                param2 = NULL;
                             }
 
                             /* Process the command. */
@@ -522,14 +548,15 @@ void commandPoll(TICKS nowTicks) {
                                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Command '%s'. Restarting JVM."), command);
                                 wrapperRestartProcess();
                             } else if (strcmpIgnoreCase(command, TEXT("STOP")) == 0) {
-                                if (params == NULL) {
+                                if (param1 == NULL) {
                                     exitCode = 0;
                                 } else {
-                                    exitCode = _ttoi(params);
+                                    exitCode = _ttoi(param1);
                                 }
                                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Command '%s'. Shutting down with exit code %d."), command, exitCode);
 
-                                wrapperStopProcess(exitCode);
+                                /* Always force the shutdown as this is an external event. */
+                                wrapperStopProcess(exitCode, TRUE);
                                 
                                 /* To make sure that the JVM will not be restarted for any reason,
                                  *  start the Wrapper shutdown process as well. */
@@ -554,24 +581,24 @@ void commandPoll(TICKS nowTicks) {
                             } else if ((strcmpIgnoreCase(command, TEXT("CONSOLE_LOGLEVEL")) == 0) ||
                                     (strcmpIgnoreCase(command, TEXT("LOGFILE_LOGLEVEL")) == 0) ||
                                     (strcmpIgnoreCase(command, TEXT("SYSLOG_LOGLEVEL")) == 0)) {
-                                if (params == NULL) {
+                                if (param1 == NULL) {
                                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s' is missing its log level."), command);
                                 } else {
-                                    logLevel = getLogLevelForName(params);
+                                    logLevel = getLogLevelForName(param1);
                                     if (logLevel == LEVEL_UNKNOWN) {
-                                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s' specified an unknown log level: '%'"), command, params);
+                                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s' specified an unknown log level: '%'"), command, param1);
                                     } else {
                                         oldLowLogLevel = getLowLogLevel();
 
                                         if (strcmpIgnoreCase(command, TEXT("CONSOLE_LOGLEVEL")) == 0) {
                                             setConsoleLogLevelInt(logLevel);
-                                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Command '%s'. Set console log level to '%s'."), command, params);
+                                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Command '%s'. Set console log level to '%s'."), command, param1);
                                         } else if (strcmpIgnoreCase(command, TEXT("LOGFILE_LOGLEVEL")) == 0) {
                                             setLogfileLevelInt(logLevel);
-                                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Command '%s'. Set log file log level to '%s'."), command, params);
+                                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Command '%s'. Set log file log level to '%s'."), command, param1);
                                         } else if (strcmpIgnoreCase(command, TEXT("SYSLOG_LOGLEVEL")) == 0) {
                                             setSyslogLevelInt(logLevel);
-                                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Command '%s'. Set syslog log level to '%s'."), command, params);
+                                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Command '%s'. Set syslog log level to '%s'."), command, param1);
                                         } else {
                                             /* Shouldn't get here. */
                                             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s' lead to an unexpected state."), command);
@@ -592,7 +619,7 @@ void commandPoll(TICKS nowTicks) {
                                     (strcmpIgnoreCase(command, TEXT("CPU_OUTPUT")) == 0) ||
                                     (strcmpIgnoreCase(command, TEXT("TIMER_OUTPUT")) == 0) ||
                                     (strcmpIgnoreCase(command, TEXT("SLEEP_OUTPUT")) == 0)) {
-                                flag = ((params != NULL) && (strcmpIgnoreCase(params, TEXT("TRUE")) == 0));
+                                flag = ((param1 != NULL) && (strcmpIgnoreCase(param1, TEXT("TRUE")) == 0));
                                 if (strcmpIgnoreCase(command, TEXT("LOOP_OUTPUT")) == 0) {
                                     wrapperData->isLoopOutputEnabled = flag;
                                 } else if (strcmpIgnoreCase(command, TEXT("STATE_OUTPUT")) == 0) {
@@ -618,6 +645,47 @@ void commandPoll(TICKS nowTicks) {
                                 } else {
                                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s'.  Tests disabled."), command);
                                 }
+                            } else if (strcmpIgnoreCase(command, TEXT("PAUSE_THREAD")) == 0) {
+                                if (wrapperData->commandFileTests) {
+                                    if (param2 == NULL) {
+                                        pauseTime = -1;
+                                    } else {
+                                        pauseTime = __max(0, __min(3600, _ttoi(param2)));
+                                    }
+                                    if (strcmpIgnoreCase(param1, TEXT("MAIN")) == 0) {
+                                        wrapperData->pauseThreadMain = pauseTime;
+                                    } else if (strcmpIgnoreCase(param1, TEXT("TIMER")) == 0) {
+                                        wrapperData->pauseThreadTimer = pauseTime;
+                                    } else if (strcmpIgnoreCase(param1, TEXT("JAVAIO")) == 0) {
+                                        wrapperData->pauseThreadJavaIO = pauseTime;
+                                    } else {
+                                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s'.  Enqueue request to pause unknown thread."), command);
+                                        pauseTime = 0;
+                                    }
+                                    if (pauseTime > 0) {
+                                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s'.  Enqueue request to pause %s thread for %d seconds..."), command, param1, pauseTime);
+                                    } else if (pauseTime < 0) {
+                                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s'.  Enqueue request to pause %s thread indefinitely..."), command, param1);
+                                    }
+                                } else {
+                                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s'.  Tests disabled."), command);
+                                }
+                            } else if (strcmpIgnoreCase(command, TEXT("PAUSE_LOGGER")) == 0) {
+                                if (wrapperData->commandFileTests) {
+                                    if (param1 == NULL) {
+                                        pauseTime = -1;
+                                    } else {
+                                        pauseTime = __max(0, __min(3600, _ttoi(param1)));
+                                    }
+                                    if (pauseTime > 0) {
+                                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s'.  Enqueue request to pause logger for %d seconds..."), command, pauseTime);
+                                    } else {
+                                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s'.  Enqueue request to pause logger indefinitely..."), command);
+                                    }
+                                    setPauseTime(pauseTime);
+                                } else {
+                                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Command '%s'.  Tests disabled."), command);
+                                }
                             } else if (strcmpIgnoreCase(command, TEXT("ACCESS_VIOLATION")) == 0) {
                                 if (wrapperData->commandFileTests) {
                                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL, TEXT("Command '%s'.  Intentionally causing an Access Violation in Wrapper..."), command);
@@ -756,7 +824,8 @@ void wStatePausing(TICKS nowTicks) {
                 (wrapperData->jState == WRAPPER_JSTATE_STOPPED) ||
                 (wrapperData->jState == WRAPPER_JSTATE_KILLING) ||
                 (wrapperData->jState == WRAPPER_JSTATE_KILL) ||
-                (wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK)) {
+                (wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK) ||
+                (wrapperData->jState == WRAPPER_JSTATE_DOWN_FLUSH)) {
                 /* In the process of stopping the JVM. */
             } else {
                 /* The JVM needs to be stopped, start that process. */
@@ -884,6 +953,7 @@ void wStateStopped(TICKS nowTicks) {
  */
 void jStateDownClean(TICKS nowTicks, int nextSleep) {
     TCHAR onExitParamBuffer[16 + 10 + 1];
+    const TCHAR *onExitAction;
     int startupDelay;
     int restartMode;
 
@@ -899,6 +969,11 @@ void jStateDownClean(TICKS nowTicks, int nextSleep) {
             /* A JVM needs to be launched. */
             restartMode = wrapperData->restartRequested;
             wrapperData->restartRequested = WRAPPER_RESTART_REQUESTED_NO;
+            wrapperData->stoppedPacketReceived = FALSE;
+            wrapperData->restartPacketReceived = FALSE;
+            if (wrapperData->isDebugging) {
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG, TEXT("Preparing to restart with mode %d."), restartMode);
+            }
 
             /* Depending on the number of restarts to date, decide how to handle the (re)start. */
             if (wrapperData->jvmRestarts > 0) {
@@ -990,7 +1065,9 @@ void jStateDownClean(TICKS nowTicks, int nextSleep) {
             /* The JVM is down, but a restart has not yet been requested.
              *   See if the user has registered any events for the exit code. */
             _sntprintf(onExitParamBuffer, 16 + 10 + 1, TEXT("wrapper.on_exit.%d"), wrapperData->exitCode);
-            if (checkPropertyEqual(properties, onExitParamBuffer, getStringProperty(properties, TEXT("wrapper.on_exit.default"), TEXT("shutdown")), TEXT("restart"))) {
+            
+            onExitAction = getStringProperty(properties, onExitParamBuffer, getStringProperty(properties, TEXT("wrapper.on_exit.default"), TEXT("shutdown")));
+            if (strcmpIgnoreCase(onExitAction, TEXT("restart")) == 0) {
                 /* We want to restart the JVM. */
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
                     TEXT("on_exit trigger matched.  Restarting the JVM.  (Exit code: %d)"), wrapperData->exitCode);
@@ -998,7 +1075,7 @@ void jStateDownClean(TICKS nowTicks, int nextSleep) {
                 wrapperData->restartRequested = WRAPPER_RESTART_REQUESTED_CONFIGURED;
 
                 /* Fall through, the restart will take place on the next loop. */
-            } else if (checkPropertyEqual(properties, onExitParamBuffer, getStringProperty(properties, TEXT("wrapper.on_exit.default"), TEXT("shutdown")), TEXT("pause"))) {
+            } else if (strcmpIgnoreCase(onExitAction, TEXT("pause")) == 0) {
                 /* We want to pause the JVM. */
                 if (wrapperData->pausable) {
                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
@@ -1011,6 +1088,12 @@ void jStateDownClean(TICKS nowTicks, int nextSleep) {
                 }
             } else {
                 /* We want to stop the Wrapper. */
+                
+                if (strcmpIgnoreCase(onExitAction, TEXT("shutdown")) != 0) {
+                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Encountered an unexpected value for configuration property %s=%s.  Resolving to %s."),
+                        onExitParamBuffer, onExitAction, TEXT("SHUTDOWN"));
+                }
+                
                 wrapperSetWrapperState(WRAPPER_WSTATE_STOPPING);
             }
         }
@@ -1032,7 +1115,9 @@ void jStateDownClean(TICKS nowTicks, int nextSleep) {
                  *  Normally, this would result in the service stopping from the paused
                  *  state, but it is possible that an exit code is registered. Check them. */
                 _sntprintf(onExitParamBuffer, 16 + 10 + 1, TEXT("wrapper.on_exit.%d"), wrapperData->exitCode);
-                if (checkPropertyEqual(properties, onExitParamBuffer, getStringProperty(properties, TEXT("wrapper.on_exit.default"), TEXT("shutdown")), TEXT("restart"))) {
+                
+                onExitAction = getStringProperty(properties, onExitParamBuffer, getStringProperty(properties, TEXT("wrapper.on_exit.default"), TEXT("shutdown")));
+                if (strcmpIgnoreCase(onExitAction, TEXT("restart")) == 0) {
                     /* We want to restart the JVM.   But not now.  Let the user know. */
                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
                         TEXT("on_exit trigger matched.  Service is paused, will restart the JVM when resumed.  (Exit code: %d)"), wrapperData->exitCode);
@@ -1041,13 +1126,19 @@ void jStateDownClean(TICKS nowTicks, int nextSleep) {
                     wrapperData->restartRequested = WRAPPER_RESTART_REQUESTED_CONFIGURED;
 
                     /* Fall through, the restart will take place once the service is resumed. */
-                } else if (checkPropertyEqual(properties, onExitParamBuffer, getStringProperty(properties, TEXT("wrapper.on_exit.default"), TEXT("shutdown")), TEXT("pause"))) {
+                } else if (strcmpIgnoreCase(onExitAction, TEXT("pause")) == 0) {
                     /* We are paused as expected. */
 
                     /* Make sure we are setup to restart when the Wrapper is resumed later. */
                     wrapperData->restartRequested = WRAPPER_RESTART_REQUESTED_CONFIGURED;
                 } else {
                     /* We want to stop the Wrapper. */
+                    
+                    if (strcmpIgnoreCase(onExitAction, TEXT("shutdown")) != 0) {
+                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT("Encountered an unexpected value for configuration property %s=%s.  Resolving to %d."),
+                            onExitParamBuffer, onExitAction, TEXT("SHUTDOWN"));
+                    }
+                    
                     wrapperSetWrapperState(WRAPPER_WSTATE_STOPPING);
                 }
             }
@@ -1156,10 +1247,10 @@ void jStateLaunchDelay(TICKS nowTicks, int nextSleep) {
             /* Generate a unique key to use when communicating with the JVM */
             wrapperBuildKey();
 
-            /* Check the backend server socket to make sure it has been initialized.
+            /* Check the backend server to make sure it has been initialized.
              *  This is needed so we can pass its port as part of the java command. */
-            if (!wrapperCheckServerSocket(TRUE)) {
-                /* The socket is not up.  An error should have been reported.  But this means we
+            if (!wrapperCheckServerBackend(TRUE)) {
+                /* The backend is not up.  An error should have been reported.  But this means we
                  *  are unable to continue. */
                 wrapperSetWrapperState(WRAPPER_WSTATE_STOPPING);
                 wrapperData->exitCode = 1;
@@ -1239,9 +1330,6 @@ void jStateLaunch(TICKS nowTicks, int nextSleep) {
         (wrapperData->wState == WRAPPER_WSTATE_STARTED) ||
         (wrapperData->wState == WRAPPER_WSTATE_RESUMING)) {
 
-        /* Always make sure that the orphan flag is reset before launching a new JVM. */
-        wrapperData->isJVMOrphaned = FALSE;
-
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Launching a JVM..."));
 
         wrapperExecute();
@@ -1400,9 +1488,7 @@ void jStateStarted(TICKS nowTicks, int nextSleep) {
     if (nextSleep && (wrapperGetProcessStatus(nowTicks, FALSE) == WRAPPER_PROCESS_DOWN)) {
         /* The process is gone.  Restart it. (Handled and logged) */
     } else {
-        if (wrapperData->isJVMOrphaned) {
-            /* The JVM requested that it be orphaned so it is no longer desirable to ping or expect ping responses.  Do nothing.*/
-        } else if (wrapperData->jStateTimeoutTicksSet && (wrapperGetTickAgeSeconds(wrapperData->jStateTimeoutTicks, nowTicks) >= 0)) {
+        if (wrapperData->jStateTimeoutTicksSet && (wrapperGetTickAgeSeconds(wrapperData->jStateTimeoutTicks, nowTicks) >= 0)) {
             /* Have we waited too long already.  The jStateTimeoutTicks is reset each time a ping
              *  response is received from the JVM. */
             if (wrapperData->debugJVM) {
@@ -1468,7 +1554,6 @@ void jStateStop(TICKS nowTicks, int nextSleep) {
         wrapperProtocolFunction(WRAPPER_MSG_STOP, NULL);
 
         /* Allow up to 5 + <shutdownTimeout> seconds for the application to stop itself. */
-        /* Already in this state. */
         if (wrapperData->shutdownTimeout > 0) {
             wrapperSetJavaState(WRAPPER_JSTATE_STOPPING, nowTicks, 5 + wrapperData->shutdownTimeout);
         } else {
@@ -1603,7 +1688,7 @@ void jStateKill(TICKS nowTicks, int nextSleep) {
 /**
  * WRAPPER_JSTATE_DOWN_CHECK
  * The JVM process currently does not exist but we still need to clean up.
- *  Once we have cleaned up, we will switch to the DOWN_CLEAN state.
+ *  Once we have cleaned up, we will switch to the DOWN_FLUSH state.
  *
  * nowTicks: The tick counter value this time through the event loop.
  * nextSleep: Flag which is used to determine whether or not the state engine
@@ -1612,6 +1697,31 @@ void jStateKill(TICKS nowTicks, int nextSleep) {
  *            function will be called again immediately.
  */
 void jStateDownCheck(TICKS nowTicks, int nextSleep) {
+    wrapperSetJavaState(WRAPPER_JSTATE_DOWN_FLUSH, nowTicks, -1);
+}
+
+/**
+ * WRAPPER_JSTATE_DOWN_FLUSH
+ * The JVM process currently does not exist but we still need to flush all of its output.
+ *  Once we have flushed and processed everything, we will switch to the DOWN_CLEAN state.
+ *
+ * nowTicks: The tick counter value this time through the event loop.
+ * nextSleep: Flag which is used to determine whether or not the state engine
+ *            will be sleeping before then next time through the loop.  It
+ *            may make sense to avoid certain actions if it is known that the
+ *            function will be called again immediately.
+ */
+void jStateDownFlush(TICKS nowTicks, int nextSleep) {
+    /* Always proceed after a single cycle. */
+    /* TODO - Look into ways of reliably detecting when the backend and stdout piles are closed. */
+    
+    /* Always close the backend here to make sure we are ready for the next JVM.
+     * In normal cases, the backend will have already been closed, but if the JVM
+     *  crashed or the Wrapper thread was delayed, then it is possible that it is
+     *  still open at this point. */
+    wrapperProtocolClose();
+    
+    /* We are now down and clean. */
     wrapperSetJavaState(WRAPPER_JSTATE_DOWN_CLEAN, nowTicks, -1);
 }
 
@@ -1678,6 +1788,7 @@ void wrapperEventLoop() {
     wrapperData->commandTimeoutTicks = lastCycleTicks;
     wrapperData->memoryOutputTimeoutTicks = lastCycleTicks;
     wrapperData->cpuOutputTimeoutTicks = lastCycleTicks;
+    wrapperData->pageFaultOutputTimeoutTicks = lastCycleTicks;
     wrapperData->logfileInactivityTimeoutTicks = lastCycleTicks;
 
     if (wrapperData->isDebugging) {
@@ -1719,6 +1830,11 @@ void wrapperEventLoop() {
         }
         maintainLogger();
         
+        if (wrapperData->pauseThreadMain) {
+            wrapperPauseThread(wrapperData->pauseThreadMain, TEXT("main"));
+            wrapperData->pauseThreadMain = 0;
+        }
+        
         /* After we maintain the logger, see if there were any signals trapped. */
 #ifdef WIN32
         wrapperMaintainControlCodes();
@@ -1730,16 +1846,18 @@ void wrapperEventLoop() {
         wrapperCheckConsoleWindows();
 #endif
 
-        /* Check the stout pipe of the child process. */
-        if (wrapperData->isLoopOutputEnabled) {
-            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("    Loop: process JVM output"));
-        }
-        if (wrapperReadChildOutput()) {
-            if (wrapperData->isDebugging) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG,
-                    TEXT("Pause reading child process output to share cycles."));
+        if (!wrapperData->useJavaIOThread) {
+            /* Check the stout pipe of the child process. */
+            if (wrapperData->isLoopOutputEnabled) {
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("    Loop: process JVM output"));
+            }
+            if (wrapperReadChildOutput()) {
+                if (wrapperData->isDebugging) {
+                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG,
+                        TEXT("Pause reading child process output to share cycles."));
+                }
+                nextSleep = FALSE;
             }
-            nextSleep = FALSE;
         }
 
         /* Check for incoming data packets. */
@@ -1747,7 +1865,7 @@ void wrapperEventLoop() {
             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("    Loop: process socket"));
         }
         /* Don't bother processing the socket if we are shutting down and the JVM is down. */
-        if (((wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK) || (wrapperData->jState == WRAPPER_JSTATE_DOWN_CLEAN)) &&
+        if (((wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK) || (wrapperData->jState == WRAPPER_JSTATE_DOWN_FLUSH) || (wrapperData->jState == WRAPPER_JSTATE_DOWN_CLEAN)) &&
             ((wrapperData->wState == WRAPPER_WSTATE_STOPPING) || (wrapperData->wState == WRAPPER_WSTATE_STOPPED))) {
             /* Skip socket processing. */
         } else {
@@ -1797,6 +1915,16 @@ void wrapperEventLoop() {
             }
         }
 
+#ifdef WIN32
+        /* Log PageFault info. */
+        if (wrapperData->isPageFaultOutputEnabled) {
+            if (wrapperTickExpired(nowTicks, wrapperData->pageFaultOutputTimeoutTicks)) {
+                wrapperDumpPageFaultUsage();
+                wrapperData->pageFaultOutputTimeoutTicks = wrapperAddToTicks(nowTicks, wrapperData->pageFaultOutputInterval);
+            }
+        }
+#endif
+
         /* Test the activity of the logfile. */
         activity = getLogfileActivity();
         if (activity != lastLogfileActivity) {
@@ -1827,25 +1955,23 @@ void wrapperEventLoop() {
         if (wrapperData->isStateOutputEnabled) {
             if (wrapperData->jStateTimeoutTicksSet) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                           TEXT("    Ticks=%08x, WrapperState=%s, JVMState=%s JVMStateTimeoutTicks=%08x (%ds), Exit=%s, RestartMode=%d, Orphan=%s"),
+                           TEXT("    Ticks=%08x, WrapperState=%s, JVMState=%s JVMStateTimeoutTicks=%08x (%ds), Exit=%s, RestartMode=%d"),
                            nowTicks,
                            wrapperGetWState(wrapperData->wState),
                            wrapperGetJState(wrapperData->jState),
                            wrapperData->jStateTimeoutTicks,
                            wrapperGetTickAgeSeconds(nowTicks, wrapperData->jStateTimeoutTicks),
                            (wrapperData->exitRequested ? TEXT("true") : TEXT("false")),
-                           wrapperData->restartRequested,
-                           (wrapperData->isJVMOrphaned ? TEXT("true") : TEXT("false")));
+                           wrapperData->restartRequested);
             } else {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                           TEXT("    Ticks=%08x, WrapperState=%s, JVMState=%s JVMStateTimeoutTicks=%08x (N/A), Exit=%s, RestartMode=%d, Orphan=%s"),
+                           TEXT("    Ticks=%08x, WrapperState=%s, JVMState=%s JVMStateTimeoutTicks=%08x (N/A), Exit=%s, RestartMode=%d"),
                            nowTicks,
                            wrapperGetWState(wrapperData->wState),
                            wrapperGetJState(wrapperData->jState),
                            wrapperData->jStateTimeoutTicks,
                            (wrapperData->exitRequested ? TEXT("true") : TEXT("false")),
-                           wrapperData->restartRequested,
-                           (wrapperData->isJVMOrphaned ? TEXT("true") : TEXT("false")));
+                           wrapperData->restartRequested);
             }
         }
 
@@ -1877,15 +2003,24 @@ void wrapperEventLoop() {
                 (wrapperData->jState == WRAPPER_JSTATE_STOPPED) ||
                 (wrapperData->jState == WRAPPER_JSTATE_KILLING) ||
                 (wrapperData->jState == WRAPPER_JSTATE_KILL) ||
-                (wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK)) {
+                (wrapperData->jState == WRAPPER_JSTATE_DOWN_CHECK) ||
+                (wrapperData->jState == WRAPPER_JSTATE_DOWN_FLUSH)) {
                 /* The JVM is already being stopped, so nothing else needs to be done. */
             } else {
                 /* The JVM should be running or is in the process of launching, so it needs to be stopped. */
                 if (wrapperGetProcessStatus(nowTicks, FALSE) == WRAPPER_PROCESS_DOWN) {
                     /* The process is gone.  (Handled and logged) */
 
-                    /* We never want to restart here. */
-                    wrapperData->restartRequested = WRAPPER_RESTART_REQUESTED_NO;
+                    if (wrapperData->restartPacketReceived) {
+                        /* The restart packet was received.  If we are here then it was delayed,
+                         *  but it means that we do want to restart. */
+                    } else {
+                        /* We never want to restart here. */
+                        wrapperData->restartRequested = WRAPPER_RESTART_REQUESTED_NO;
+                        if (wrapperData->isDebugging) {
+                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG, TEXT("Reset the restart flag."));
+                        }
+                    }
                 } else {
                     /* JVM is still up.  Try asking it to shutdown nicely. */
                     if (wrapperData->isDebugging) {
@@ -1999,6 +2134,10 @@ void wrapperEventLoop() {
             jStateDownCheck(nowTicks, nextSleep);
             break;
 
+        case WRAPPER_JSTATE_DOWN_FLUSH:
+            jStateDownFlush(nowTicks, nextSleep);
+            break;
+
         default:
             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("Unknown jState=%d"), wrapperData->jState);
             break;
