@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2010 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -31,10 +31,12 @@
 #include <string.h>
 #include <errno.h>
 #include <stdlib.h>
-
+#include <fcntl.h>
 #ifdef WIN32
  #include <windows.h>
  #include <tchar.h>
+ #include <io.h>
+ #define dup2 _dup2
 #endif
 #include "wrapper_i18n.h"
 #include "wrapperinfo.h"
@@ -47,6 +49,13 @@ int controlEventQueue[CONTROL_EVENT_QUEUE_SIZE];
 int controlEventQueueLastReadIndex = 0;
 int controlEventQueueLastWriteIndex = 0;
 
+
+/** Flag to keep track of whether StdOut has been redirected. */
+int redirectedStdOut = FALSE;
+
+/** Flag to keep track of whether StdErr has been redirected. */
+int redirectedStdErr = FALSE;
+
 const char utf8ClassJavaLangString[] = {106, 97,118, 97, 47, /* java/ */
                                         108, 97,110,103, 47, /* lang/ */
                                         83,116,114,105,110,103, 0}; /* "java/lang/String" */
@@ -171,7 +180,7 @@ jstring JNU_NewStringNative(JNIEnv *env, const TCHAR *strW) {
     len = _tcslen(strW);
     if (len > 0) {
  #ifdef WIN32
-        size = WideCharToMultiByte(CP_ACP, 0, strW, -1, NULL, 0, NULL, NULL);
+        size = WideCharToMultiByte(CP_UTF8, 0, strW, -1, NULL, 0, NULL, NULL);
         if (size == 0) {
             /* Failed. */
             _tprintf(TEXT("WrapperJNI Warn: Failed to convert string \"%s\": %s\n"), strW, GetLastError()); fflush(NULL);
@@ -182,7 +191,10 @@ jstring JNU_NewStringNative(JNIEnv *env, const TCHAR *strW) {
             throwOutOfMemoryError(env, TEXT("JNSN1"));
             return NULL;
         }
-        WideCharToMultiByte(CP_ACP, 0, strW, -1, msgMB, size, NULL, NULL);
+        WideCharToMultiByte(CP_UTF8, 0, strW, -1, msgMB, size, NULL, NULL);
+        result = (*env)->NewStringUTF(env, msgMB);
+        free(msgMB);
+        return result;
  #else
         size = wcstombs(NULL, strW, 0) + 1;
         msgMB = malloc(sizeof(char) * size);
@@ -207,6 +219,11 @@ jstring JNU_NewStringNative(JNIEnv *env, const TCHAR *strW) {
     result = NULL;
     if ((*env)->EnsureLocalCapacity(env, 2) < 0) {
         throwOutOfMemoryError(env, TEXT("JNSN4"));
+#ifdef UNICODE
+        if (msgMB) {
+            free(msgMB);
+        }
+#endif
         return NULL; /* out of memory error */
     }
     len = strlen(msgMB);
@@ -312,6 +329,7 @@ TCHAR *JNU_GetStringNativeChars(JNIEnv *env, jstring jstr) {
     size = MultiByteToWideChar(CP_OEMCP, 0, result, -1, NULL, 0);
     tresult = malloc(size*sizeof(LPWSTR));
     if (!tresult) {
+        free(result);
         throwOutOfMemoryError(env, TEXT("GSNC3"));
         return NULL;
     }
@@ -323,6 +341,7 @@ TCHAR *JNU_GetStringNativeChars(JNIEnv *env, jstring jstr) {
     size *= sizeof(TCHAR);
     tresult = malloc(size);
     if (!tresult) {
+        free(result);
         throwOutOfMemoryError(env, TEXT("GSNC3"));
         return NULL;
     }
@@ -456,10 +475,12 @@ int getSystemProperty(JNIEnv *env, const TCHAR *propertyName, TCHAR **propertyVa
                                 throwOutOfMemoryError(env, TEXT("GSP1"));
                                 result = TRUE;
                                 } else {
-                                _tcscpy(*propertyValue, keyChars);
+                                _tcsncpy(*propertyValue, keyChars, _tcslen(keyChars) + 1);
                                 result = FALSE;
                             }
-                            free(keyChars);
+                            if (keyChars) {
+                                free(keyChars);
+                            }
                         } else {
                             /* Exception Thrown */
                             result = TRUE;
@@ -472,8 +493,8 @@ int getSystemProperty(JNIEnv *env, const TCHAR *propertyName, TCHAR **propertyVa
                             if (!*propertyValue) {
                                 throwOutOfMemoryError(env, TEXT("GSP2"));
                                 result = TRUE;
-                                } else {
-                                strcpy((char*)*propertyValue, (char*)keyChars);
+                            } else {
+                                strncpy((char*)*propertyValue, (char*)keyChars, strlen((char*)keyChars) + 1);
                                 result = FALSE;
                             }
                             (*env)->ReleaseStringUTFChars(env, jStringKeyValue, (const char *)keyChars);
@@ -508,7 +529,44 @@ int getSystemProperty(JNIEnv *env, const TCHAR *propertyName, TCHAR **propertyVa
  * Do common initializaion.
  */
 void initCommon(JNIEnv *env, jclass jClassWrapperManager) {
+    TCHAR* outfile, *errfile;
+    int outfd, errfd, mode, options;
+
+#ifdef WIN32
+    mode = _S_IWRITE;
+    options = _O_WRONLY | _O_APPEND | _O_CREAT;
+#else
+    mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
+    options = O_WRONLY | O_APPEND | O_CREAT;
+#endif
     initUTF8Strings(env);
+
+    if (getSystemProperty(env, TEXT("wrapper.java.errfile"), &errfile, FALSE)) {
+        /* Failed */
+        return;
+    }
+    if (errfile) {
+        _ftprintf(stderr, TEXT("WrapperJNI: Redirecting %s to file %s...\n"), TEXT("StdErr"), errfile); fflush(NULL);
+        if (((errfd = _topen(errfile, options, mode)) == -1) || (dup2(errfd, STDERR_FILENO) == -1)) {
+            _ftprintf(stderr, TEXT("WrapperJNI: Failed to redirect %s to file %s  (Err: %s)\n"), TEXT("StdErr"), errfile, getLastErrorText()); fflush(NULL);
+            return;
+        } else {
+            redirectedStdErr = TRUE;
+        }
+    }
+    if (getSystemProperty(env, TEXT("wrapper.java.outfile"), &outfile, FALSE)) {
+        /* Failed */
+        return;
+    }
+    if (outfile) {
+        _tprintf(TEXT("WrapperJNI: Redirecting %s to file %s...\n"), TEXT("StdOut"), outfile); fflush(NULL);
+        if (((outfd = _topen(outfile, options, mode)) == -1) || (dup2(outfd, STDOUT_FILENO) == -1)) {
+            _tprintf(TEXT("WrapperJNI: Failed to redirect %s to file %s  (Err: %s)\n"), TEXT("StdOut"), errfile, getLastErrorText()); fflush(NULL);
+            return;
+        } else {
+            redirectedStdOut = TRUE;
+        }
+    }
 }
 
 void throwThrowable(JNIEnv *env, char *throwableClassName, const TCHAR *lpszFmt, ...) {
@@ -735,6 +793,7 @@ Java_org_tanukisoftware_wrapper_WrapperManager_nativeGetControlEvent(JNIEnv *env
         fflush(NULL);
 #endif
     }
+
     if (wrapperReleaseControlEventQueue()) {
         /* Failed.  Should have been reported. */
         return event;
